#ifndef MAKEDEPEND
#include <cmath>
#endif

#include "expr.h"

class expr_math : public expr
{
private:
	expr_floattype (*func)(expr_floattype);
protected:
	exprp arg;
public:
	expr_math(const expr_math &s) : expr(s)
	{
		func = s.func;
		arg = s.arg;
	}
	expr_math(string &input, expr_floattype (*f)(expr_floattype)) : func(f)
	{
		arg = expr::load(input, prio_high)->optimize();
		if(!arg)
			valid = false;
		else if(arg->type == str)
		{
			valid = false;
			arg = NULL;
		}
		else
		{
			type = num;
			if(arg->isconst)isconst = true;
		}
	}
 	virtual const char *fname() const { return NULL; };
	virtual const string getname() const
	{
		return string(fname()) + '(' + arg->getname() + ')';
	}
	virtual exprp optimize()
	{
		arg = arg->optimize();
		if(arg->isconst && arg->type == num)
		{
			expr_floattype d = func(arg->getnum());
			return new expr_number(d);
		}
		return this;
	}
	virtual exprp calculate(calculate_args) const
	{
		exprp tmp = arg->calculate(calculate_call);
		if(tmp->isconst
		#ifndef DUD2
		&& tmp->type!=notsure
		#endif
		)
		{
			expr_floattype d = func(tmp->getnum());
			return new expr_number(d);
		}
		return new expr_math(*this);
	}
};


#if expr_need_l
 #define math_expr_class(func) \
 class expr_##func : public expr_math \
 { public: \
 	expr_##func(string &input) : expr_math(input, func##l) { } \
 	virtual const char *fname() const { return #func; } \
 };
 #define math_atan2_fun atan2l
#else
 #define math_expr_class(func) \
 class expr_##func : public expr_math \
 { public: \
    expr_##func(string &input) : expr_math(input, func) { fname=#func; } \
 	virtual const char *fname() const { return #func; } \
 };
 #define math_atan2_fun atan2
#endif

class expr_atan2 : public expr
{
public:
    exprp a1, a2;
	expr_atan2(const expr_atan2 &s) : expr(s)
	{
		a1 = s.a1;
		a2 = s.a2;
	}
    expr_atan2(string &input)
    {
      blank(input);
      if(input[0] != '(') valid = false;
      else {
        getchar(input);
        a1 = expr::load(input, prio_arglist);
        if(!a1)valid = false;
        else {
          blank(input);
          if(input[0] == ',') getchar(input);
          a2 = expr::load(input, prio_arglist);
          if(!a2)valid = false;
          blank(input);
          if(input[0] != ')')valid = false;
          else getchar(input);
    } } }
	virtual const string getname() const
	{
		return "atan2(" + a1->getname() + ", " + a2->getname() + ')';
	}
	virtual exprp optimize()
	{
		a1 = a1->optimize();
		a2 = a2->optimize();
		if(a1->isconst && a1->type == num && a2->isconst && a2->type == num)
		{
			expr_floattype d = math_atan2_fun(a1->getnum(), a2->getnum());
			return new expr_number(d);
		}
		return this;
	}
	virtual exprp calculate(calculate_args) const
	{
		exprp t1 = a1->calculate(calculate_call),
		      t2 = a2->calculate(calculate_call);
		if(t1->isconst && t2->isconst
		&& t1->type!=notsure
		&& t2->type!=notsure)
		{
			expr_floattype d = math_atan2_fun(t1->getnum(), t2->getnum());
			return new expr_number(d);
		}
		return new expr_atan2(*this);
	}
};

math_expr_class(sin)
math_expr_class(cos)
math_expr_class(tan)
math_expr_class(sqrt)
math_expr_class(sinh)
math_expr_class(cosh)
math_expr_class(tanh)
math_expr_class(exp)
math_expr_class(log)
math_expr_class(log10)
math_expr_class(atan)
math_expr_class(asin)
math_expr_class(acos)
math_expr_class(ceil)
math_expr_class(floor)

#undef math_expr_class
